#if defined COUP_OAS
SUBROUTINE oas_cos_define
!
!**** *oas_cos_define*  - Define grids etc for OASIS
!
!     Purpose. 
!     --------
!        Define grid and field information for atmosphere
!        exchange between AGCM, OGCM and COUPLER. (OASIS3 software)
!
!**   Interface.
!     ----------
!       *CALL*  *oas_cos_define*
!
!     Input:
!     -----
!
!     Output:
!     ------
!
!     Method:
!     ------
!        
!     Externals:
!     ---------
!       prism_init, prism_init_comp, prism_get_localcomm : prism library
!
!     Author:
!     -------
!       R. Redler, NEC Laboratories Europe
!       K. Mogensen, ECMWF
!
!     Modifications.
!     --------------
!       E. Maisonnave : adapted from NEMO to COSMO
!       A. Dobler: adapted to OASIS3
!       E. Maisonnave : adapted to COSMO
!       Naveed Akhtar: adapted to OASIS3-MCT2
!       2011-11 J. Brauch: calculate area
!       2012-08: unified interface for OASIS - COSMO/CLM, Will/Davin/Brauch/Weiher
!       2013-06: extension for staggered variables for echam-coupling Weiher
!    2017 Naveed Akhtar implemented coupling TRIP river routing scheme
!    2017 Naveed Akhtar implemented E-P of Black Sea coupling
USE data_parameters,  ONLY : &
  ireals,                 &
  iintegers

USE data_runcontrol,  ONLY : &
  hstart,                 &
  nuspecif

USE oas_cos_vardef

#if ! defined OASIS3MCT
  USE mod_prism_grids_writing
  USE mod_prism_def_partition_proto
#else
  USE mod_oasis_namcouple    ! OASIS3MCT namcouple variables: e.g. coupling time step
#endif

USE data_modelconfig, ONLY : &
  ie,                     &
  je,                     &
  dlon,                   &
  dlat,                   &
  startlon_tot,           &
  startlat_tot,           &
  raddeg,                 &
  ie_tot,                 &
  je_tot,                 &
  dt,                     &
  pollon,                 & ! longitude of the rotated north pole (in degrees, E>0)
  pollat,                 & ! latitude of the rotated north pole (in degrees, N>0)
  polgam                    ! angle between the north poles of the systems

USE data_fields,      ONLY : &
  rlat,                   &
  rlon,                   &
  fr_land

USE data_constants,   ONLY : &
  r_earth

USE data_parallel,    ONLY : &
  my_cart_id,             & ! rank of this subdomain in the cartesian communicator
  isubpos,                & ! positions of the subdomains in the total domain. Given
                            ! are the i- and the j-indices of the lower left and the
                            ! upper right grid point in the order
                            !                  i_ll, j_ll, i_ur, j_ur.
                            ! Only the interior of the domains are considered, not
                            ! the boundary lines.
  nboundlines,            & ! number of boundary lines of the domain for which
                            ! no forecast is computed = overlapping boundary
                            ! lines of the subdomains
  intbuf,                 & !
  charbuf,imp_integers
                            
USE parallel_utilities, ONLY : &
  gather_field              ! gathers the parts of a total field from all subdomains


USE utilities,          ONLY :  &
  phirot2phi,               & !
  rlarot2rla                  !


IMPLICIT NONE

!
! local variables
!

!EM 24.08.2012 OASIS3 compatibility: dummy variables
#if ! defined OASIS3MCT
  INTEGER(KIND=iintegers)         :: namflddti(1)
  CHARACTER (LEN=8)               :: namdstfld(1)
#endif

INTEGER(KIND=iintegers) :: &
  igrid,                   & ! ids returned by prism_def_grid
  iptid,                   & ! ids returned by prism_set_points 
  imskid,                  & ! ids returned by prism_set_mask
  iextent(1,3),            & !
  ioffset(1,3),            & ! 
  ji, jj, jg,              & !
  id_part,                 & !
  paral(5),                & ! OASIS3 box partition
  var_nodims(2),           & ! 
  ishape(2,2),             & ! shape of arrays passed to PSMILe
  nlei_tot, nlej_tot,      & ! upper halo limits on a global subdomain
  jih_tot, jjh_tot,        & ! global subdomain size without halo
  ierrstat

REAL(KIND=ireals)       :: &
  start_lonc,              &
  start_latc

CHARACTER(LEN=4)        :: &
  clgrd   = 'cosl'         ! name of grid denoting grid points at mass center




! local arrays
!

INTEGER(KIND=iintegers)      :: &
  imask_tot (ie_tot, je_tot)      !

REAL(KIND=ireals)            :: &
  zclo      (ie    ,je, 4),     &
  zcla      (ie    ,je, 4),     &
  zlon      (ie    ,je),        &
  zlat      (ie    ,je),        &
  zlonu     (ie    ,je),        &
  zlatu     (ie    ,je),        &
  zlonv     (ie    ,je),        &
  zlatv     (ie    ,je),        &
  zmask     (ie    ,je),        &
  zmask_tot (ie_tot,je_tot),    & 
  area_tot  (ie_tot,je_tot),    &
  zclo_tot  (ie_tot,je_tot,4),  &
  zcla_tot  (ie_tot,je_tot,4),  &
  zclou_tot (ie_tot,je_tot,4),  &
  zclau_tot (ie_tot,je_tot,4),  &
  zclov_tot (ie_tot,je_tot,4),  &
  zclav_tot (ie_tot,je_tot,4),  &
  zlon_tot  (ie_tot,je_tot),    & 
  zlat_tot  (ie_tot,je_tot),    &
  zslonu_tot(ie_tot,je_tot),    &
  zslatu_tot(ie_tot,je_tot),    &
  zslonv_tot(ie_tot,je_tot),    &
  zslatv_tot(ie_tot,je_tot),    &
  zslonu    (ie_tot,je_tot),    & ! U-direction shifted longitudes
  zslatu    (ie_tot,je_tot),    & ! U-direction shifted latitudes
  zslonv    (ie_tot,je_tot),    & ! V-direction shifted longitudes
  zslatv    (ie_tot,je_tot),    & ! V-direction shifted latitudes
  zsrlon    (ie_tot),           & ! shifted rotated longitudes
  zsrlat    (je_tot),           & ! shifted rotated latitudes
  zrlon     (ie_tot),           & ! rotated longitudes
  zrlat     (je_tot)              ! rotated latitudes

!------------------------------------------------------------------------------
!- End of header
!------------------------------------------------------------------------------
! -----------------------------------------------------------------
! ... Setup the OASIS interface
! ----------------------------------------------------------------
      
IF ( my_cart_id == 0 .AND. debug_oasis > 0 ) THEN
  WRITE(nulout,'(A)') 'Setting up OASIS3(-MCT) interface'
  CALL flush(nulout)
ENDIF
      
! Subdomain size without halo
nldi = 1  + nboundlines
nldj = 1  + nboundlines
nlei = ie - nboundlines
nlej = je - nboundlines
      
iboundstart = 0
jboundstart = 0
      
! Subdomain size without halo BUT with boundlines
IF ( isubpos(my_cart_id,1) == 1 + nboundlines ) THEN
  IF ( debug_oasis > 0 ) WRITE(nulout,*) 'first line'
  nldi = 1
  iboundstart = nboundlines
ENDIF
IF ( isubpos(my_cart_id,3) == ie_tot - nboundlines ) THEN
  IF ( debug_oasis > 0 ) WRITE(nulout,*) 'last line'
  nlei = ie
ENDIF
IF ( isubpos(my_cart_id,2) == 1 + nboundlines ) THEN
  IF ( debug_oasis > 0 ) WRITE(nulout,*) 'first column'
  nldj = 1
  jboundstart = nboundlines
ENDIF
IF ( isubpos(my_cart_id,4) == je_tot - nboundlines ) THEN
  IF ( debug_oasis > 0 ) WRITE(nulout,*) 'last column'
  nlej = je
ENDIF
CALL flush(nulout)
      
! Local extend
jih = nlei - nldi + 1
jjh = nlej - nldj + 1
      
! By default, all grid points are unmasked
  zmask(:,:) = 1.0_ireals
  WHERE (fr_land(nldi:nlei, nldj:nlej) > 0.5_ireals) zmask(nldi:nlei, nldj:nlej) = 0.0_ireals

!
! With OASIS3, all domains involved
	
	lpe_cpl = .TRUE.
	write_aux_files=1


! -----------------------------------------------------------------
! ... Writing of auxiliary files (grids.nc, masks.nc, areas.nc)
! -----------------------------------------------------------------

IF ( write_aux_files == 1 ) THEN

! -----------------------------------------------------------------
! ... Define the center points for the unstaggered variables
! -----------------------------------------------------------------

  ! conversion to degrees
  zlon = rlon * raddeg
  zlat = rlat * raddeg


! -----------------------------------------------------------------
! ... Define the corners for the unstaggered grid points
! -----------------------------------------------------------------

  ! 1: lower left corner. 2: lower right corner 
  ! 3: upper right corner 4: upper left corner
  ! (anti-clockwise, starting from the bottom left corner)
  
  start_lonc = startlon_tot - dlon * 0.5_ireals
  start_latc = startlat_tot - dlat * 0.5_ireals

  CALL constant_corners(zclo(:,:,1), zcla(:,:,1), start_lonc, start_latc)

  start_lonc = startlon_tot + dlon * 0.5_ireals
  start_latc = startlat_tot - dlat * 0.5_ireals

  CALL constant_corners(zclo(:,:,2), zcla(:,:,2), start_lonc, start_latc)

  start_lonc = startlon_tot + dlon * 0.5_ireals
  start_latc = startlat_tot + dlat * 0.5_ireals

  CALL constant_corners(zclo(:,:,3), zcla(:,:,3), start_lonc, start_latc)

  start_lonc = startlon_tot - dlon * 0.5_ireals
  start_latc = startlat_tot + dlat * 0.5_ireals

  CALL constant_corners(zclo(:,:,4), zcla(:,:,4), start_lonc, start_latc)

  zclo = zclo * raddeg
  zcla = zcla * raddeg

  IF ( debug_oasis > 0 .AND. my_cart_id == 0 ) THEN
    WRITE(nulout,*) ' zlon ', MINVAL(zlon(:,:)), MAXVAL(zlon(:,:))
    WRITE(nulout,*) ' zlatl', MINVAL(zlat(:,:)), MAXVAL(zlat(:,:))
    WRITE(nulout,*) ' zclo ll', MINVAL(zclo(:,:,1)), MAXVAL(zclo(:,:,1))
    WRITE(nulout,*) ' zclo ur', MINVAL(zclo(:,:,3)), MAXVAL(zclo(:,:,3))
    WRITE(nulout,*) ' zcla ll', MINVAL(zcla(:,:,1)), MAXVAL(zcla(:,:,1))
    WRITE(nulout,*) ' zcla ur', MINVAL(zcla(:,:,3)), MAXVAL(zcla(:,:,3))
    CALL flush(nulout)
  ENDIF

  ! Gather information
  CALL gather_field(zlon,ie,je,zlon_tot,ie_tot,je_tot,0,nerror) 
  CALL gather_field(zlat,ie,je,zlat_tot,ie_tot,je_tot,0,nerror) 
  CALL gather_field(zmask,ie,je,zmask_tot,ie_tot,je_tot,0,nerror) 
  DO jg =1, 4
    CALL gather_field(zclo(:,:,jg),ie,je,zclo_tot(:,:,jg),ie_tot,je_tot,0,nerror) 
    CALL gather_field(zcla(:,:,jg),ie,je,zcla_tot(:,:,jg),ie_tot,je_tot,0,nerror) 
  ENDDO
  IF ( debug_oasis == 1 .AND. my_cart_id == 0 ) THEN
    WRITE(nulout,*) ' gather_field for corners'
    CALL flush(nulout)
  ENDIF

! -----------------------------------------------------------------
! ... Write auxiliary files by the master processor 
! -----------------------------------------------------------------

  IF ( my_cart_id == 0 ) THEN
 
    CALL prism_start_grids_writing (write_aux_files)

    IF ( debug_oasis == 1 ) THEN
      WRITE(nulout,*) ' prism_start_grids_writing ', write_aux_files
      CALL flush(nulout)
    ENDIF

! -----------------------------------------------------------------
! ... Write grids.nc for (un-)staggered grid points
! -----------------------------------------------------------------

    CALL prism_write_grid (clgrd, ie_tot, je_tot, zlon_tot, zlat_tot) 
    CALL prism_write_corner (clgrd, ie_tot, je_tot, 4, zclo_tot, zcla_tot)


    IF ( debug_oasis == 1 ) THEN
      WRITE(nulout,*) ' prism_write_grid and prism_write_corner '
      CALL flush(nulout)
    ENDIF

! -----------------------------------------------------------------
! ... Write masks.nc for (un-)staggered grid points
! -----------------------------------------------------------------

    imask_tot=INT(zmask_tot)

    CALL prism_write_mask (clgrd, ie_tot, je_tot, imask_tot)


    IF ( debug_oasis == 1 ) THEN
      WRITE(nulout,*) ' prism_write_mask '
      CALL flush(nulout)
    ENDIF

! -----------------------------------------------------------------
! ... Calculate grid point areas and write areas.nc 
! -----------------------------------------------------------------

    CALL calculate_area (zlon_tot,zlat_tot,area_tot)
    CALL prism_write_area (clgrd, ie_tot, je_tot, area_tot)


    IF ( debug_oasis == 1 ) THEN
      WRITE(nulout,*) ' calculate_area and prism_write_area'
      CALL flush(nulout)
    ENDIF

    CALL prism_terminate_grids_writing()

  ENDIF ! my_cart_id

ENDIF ! write_aux_files

IF ( debug_oasis == 1 .AND. my_cart_id == 0) THEN
  WRITE(nulout,*) ' prism_terminate_grids_writing '
  CALL flush(nulout)
ENDIF

#if ! defined OASIS3MCT
  ! Wait for coupling result (involved PEs and non-involved PEs)
  CALL MPI_Bcast( write_aux_files, 1, MPI_INT, 0, kl_comm, nerror )
  ! Wait for OASIS signal to continue the simulation
  IF ( write_aux_files == 1 ) CALL MPI_Barrier(mpi_comm, nerror)
#endif


IF ( lpe_cpl ) THEN

! -----------------------------------------------------------------
! ... Define the partition
! -----------------------------------------------------------------

  ! whole domain decomposition can be represented by a rectangle
  paral(1) = 2  ! 2 means box partitioning
  ! Global extent in x
  paral(5) = ie_tot
  ! Upper left corner global offset
  paral(2) = isubpos(my_cart_id,1)-1-iboundstart + ( isubpos(my_cart_id,2)-1-jboundstart) * paral(5)
  ! Local extent in x
  paral(3) = jih
  ! Local extent in y
  paral(4) = jjh

  ! Compute global offsets and local extents
  IF ( debug_oasis > 0 ) THEN
    WRITE(nulout,*) 'local ', paral(3:4), my_cart_id
    WRITE(nulout,*) 'global ', paral(2), paral(5), my_cart_id
    CALL flush(nulout)
  ENDIF
  
  CALL prism_def_partition_proto( id_part, paral, nerror )

  IF ( nerror /= PRISM_Success ) THEN
    CALL prism_abort_proto (ncomp_id, 'oas_cos_define', 'Failure in prism_def_partition' )
  ENDIF

! -----------------------------------------------------------------
! ... Initialize some variables
! ----------------------------------------------------------------

  nfld_snd_oce = 0
  nfld_rcv_oce = 0

  nfld_snd_tot = 0
  nfld_rcv_tot = 0

! -----------------------------------------------------------------
! ... Define list of SENT variable names per coupling
! ----------------------------------------------------------------

  nfld = 0


    nfld_snd_oce = 8 !! changed from 1 to 5 (naveed akhtar)
    ALLOCATE ( nlev_snd_oce (nfld_snd_oce), STAT=ierrstat )
    ALLOCATE ( nam_snd_oce  (nfld_snd_oce), STAT=ierrstat )
    nlev_snd_oce = (/1,1,1,1,1,1,1,1/)
    nam_snd_oce = (/'EMPtNM','UMFLStNM','VMFLStNM','SOBStNM','NSOBStNM','EMPBLACK','S_RUNOFF','G_RUNOFF'/)
    ! nlev_snd_oce = (/1,1/)
    !nam_snd_oce =(/'S_RUNOFF','G_RUNOFF'/)  
 DO ji = 1, nfld_snd_oce
      nfld = nfld + nlev_snd_oce(ji)
ENDDO
  nfld_snd_tot = nfld
 

! -----------------------------------------------------------------
! ... Define list of RECEIVED variable names per coupling
! ----------------------------------------------------------------

  nfld = 0


    nfld_rcv_oce = 1
    ALLOCATE ( nlev_rcv_oce (nfld_rcv_oce), STAT=ierrstat )
    ALLOCATE ( nam_rcv_oce  (nfld_rcv_oce), STAT=ierrstat )
    nlev_rcv_oce = (/1/)
    nam_rcv_oce = (/'SSTfNM'/) !! UI-MCT name changed from COSSSTEM for nemo
    DO ji = 1, nfld_rcv_oce
      nfld = nfld + nlev_rcv_oce(ji)
    ENDDO
 
  ! total number of fields to be received:
  nfld_rcv_tot = nfld

  IF ( debug_oasis == 1 .AND. my_cart_id == 0) THEN
    WRITE(nulout,*) ' nfld_snd_tot=',nfld_snd_tot,' nfld_rcv_tot=',nfld_rcv_tot
    CALL flush(nulout)
  ENDIF

! -----------------------------------------------------------------
! ... write variable names for all active couplings to one structure
! ----------------------------------------------------------------

  ! Allocate memory for data exchange:
  ALLOCATE( ssnd(nfld_snd_tot), stat = nerror )
  ALLOCATE( srcv(nfld_rcv_tot), stat = nerror )
  ALLOCATE( exfld(nldi:nlei, nldj:nlej), stat = nerror )

  IF ( nerror > 0 ) THEN
    CALL prism_abort_proto( ncomp_id, 'oas_cos_define', 'Failure in allocating exfld, ssnd or srcv' )
    RETURN
  ENDIF

  ! fill ssnd with names of fields to be sent
  nfld = 0  


    DO ji = 1, nfld_snd_oce
      nfld = nfld + 1
      ssnd(nfld)%clname = TRIM(nam_snd_oce(ji))
    ENDDO

  ! fill srcv with names of fields to be received:
  nfld = 0

    DO ji = 1, nfld_rcv_oce
      nfld = nfld + 1
      srcv(nfld)%clname = TRIM(nam_rcv_oce(ji))
    ENDDO

! ----------------------------------------------------------------------------
! ... Variable selection
! ----------------------------------------------------------------------------

  ! This is still preliminary: set laction for all (this way no subsets of
  ! variables can be selected; this has to be changed yet)
  ssnd(:)%laction = .TRUE.
  srcv(:)%laction = .TRUE.

  var_nodims(1) = 2   ! Dimension number of exchanged arrays
  var_nodims(2) = 1   ! number of bundles (always 1 for OASIS3)

! -----------------------------------------------------------------
! ... Define the shape of the valid region without the halo and overlaps between CPUs
! -----------------------------------------------------------------

  ishape(1,1) = 1
  ishape(2,1) = jih
  ishape(1,2) = 1
  ishape(2,2) = jjh

! -----------------------------------------------------------------
! ... Announce variables to be sent and to be received
! -----------------------------------------------------------------

  IF ( debug_oasis == 1 .AND. my_cart_id == 0) THEN
    WRITE(nulout,*) ' ishape=',ishape,' id_part=',id_part
    CALL flush(nulout)
  ENDIF

  ! Announce variables to be sent:
  DO ji = 1, nfld_snd_tot
 IF (ssnd(ji)%laction) THEN 
   CALL prism_def_var_proto( ssnd(ji)%nid, ssnd(ji)%clname, id_part, &
      var_nodims, PRISM_Out, ishape, PRISM_REAL, nerror )
    IF ( nerror /= PRISM_Success ) CALL prism_abort_proto( ssnd(ji)%nid, &
      'oas_cos_define', 'Failure in prism_def_var for '//TRIM(ssnd(ji)%clname) )
 ENDIF
  ENDDO
      
  ! Announce variables to be received:
  DO ji = 1, nfld_rcv_tot
  IF (srcv(ji)%laction) THEN 
   CALL prism_def_var_proto( srcv(ji)%nid, srcv(ji)%clname, id_part, &
      var_nodims, PRISM_In, ishape, PRISM_REAL, nerror )
    IF ( nerror /= PRISM_Success ) CALL prism_abort_proto( srcv(ji)%nid, &
      'oas_cos_define', 'Failure in prism_def_var for '//TRIM(srcv(ji)%clname) )
  ENDIF
  ENDDO


  ! Allocate array to store received fields between two coupling steps
  ALLOCATE( frcv(ie, je, nfld_rcv_tot), stat = nerror )
  IF ( nerror > 0 ) THEN
    CALL prism_abort_proto( ncomp_id, 'oas_cos_define', 'Failure in allocating frcv' )
    RETURN
  ENDIF   

! ------------------------------------------------------------------
! ... End of definition phase (must be called by all processes including 
!     the PEs not involved in the coupling)
! ------------------------------------------------------------------
      
  CALL prism_enddef_proto( nerror )

  IF ( nerror /= PRISM_Success ) CALL prism_abort_proto ( ncomp_id, &
    'oas_cos_define', 'Failure in prism_enddef')

  IF ( debug_oasis == 1 .AND. my_cart_id == 0) THEN
    WRITE(nulout,*) ' oas_cos_define: end of definition phase'
    CALL flush(nulout)
  ENDIF

! -----------------------------------------------------------------
! ... Deallocate arrays that are not needed anymore
! -----------------------------------------------------------------


  IF (ALLOCATED (nlev_snd_oce)) THEN
    DEALLOCATE ( nlev_snd_oce, STAT=ierrstat )
    DEALLOCATE ( nam_snd_oce , STAT=ierrstat )
  ENDIF
  IF (ALLOCATED (nlev_rcv_oce)) THEN
    DEALLOCATE ( nlev_rcv_oce, STAT=ierrstat )
    DEALLOCATE ( nam_rcv_oce , STAT=ierrstat )
  ENDIF

ENDIF ! lpe_cpl 
      
CALL MPI_Barrier(kl_comm, nerror)

END SUBROUTINE oas_cos_define

   
!==============================================================================
!+ calculate corners
!------------------------------------------------------------------------------

   SUBROUTINE constant_corners(longc,latc,startlonc,startlatc)

   USE data_parameters,  ONLY : ireals, iintegers
   USE data_parallel,    ONLY :   &
                            my_cart_id,      & ! rank of this subdomain in the cartesian communicator
                            isubpos,         & ! positions of the subdomains in the total domain. Given
                                               ! are the i- and the j-indices of the lower left and the
                                               ! upper right grid point in the order
                                               !                  i_ll, j_ll, i_ur, j_ur.
                                               ! Only the interior of the domains are considered, not
                                               ! the boundary lines.
                            nboundlines        ! number of boundary lines of the domain for which
                                               ! no forecast is computed = overlapping boundary
                                               ! lines of the subdomains
   USE data_modelconfig, ONLY : ie, je, dlon, dlat, pollon, pollat, polgam, degrad
   USE utilities,          ONLY :  &
      phirot2phi,          & !
      rlarot2rla             !

   IMPLICIT NONE

   REAL (KIND=ireals),   INTENT (IN)    ::  startlonc,startlatc
   REAL (KIND=ireals),   INTENT (OUT)   ::  longc(ie,je),latc(ie,je)
   ! Local variables:
   REAL (KIND=ireals)        ::  zlats, zlons

   INTEGER (KIND=iintegers)  ::  i, j, i_td, j_td


   ! constant fields related to the grid (rlat, rlon)
   ! ---------------------------------------------------------------------

   j_td = isubpos(my_cart_id,2) - nboundlines - 1

   DO j = 1 , je
      j_td = j_td + 1
      ! cos (lat) and 1 / cos (lat)
      zlats        = startlatc + (j_td-1) * dlat

      i_td = isubpos(my_cart_id,1) - nboundlines - 1

      DO i = 1 , ie
         i_td = i_td + 1
         ! geographical latitude and longitude
         zlons  = startlonc + (i_td-1) * dlon

         IF (zlons  > 180.0) THEN
           zlons  = zlons  - 360.0_ireals
         ENDIF

         latc(i,j) = phirot2phi ( zlats , zlons , pollat, pollon, polgam) * degrad
         longc(i,j) = rlarot2rla ( zlats , zlons , pollat, pollon, polgam) * degrad

      ENDDO
   ENDDO

   END SUBROUTINE constant_corners

!==============================================================================
!+ calculate area of each grid point
!------------------------------------------------------------------------------

   SUBROUTINE calculate_area(long,lati,surface)

   USE data_parameters,  ONLY : ireals, iintegers
   USE data_fields,      ONLY : rlat, rlon
   USE data_constants,   ONLY : r_earth
   USE data_modelconfig, ONLY : ie_tot, je_tot, raddeg

   IMPLICIT NONE

   REAL (KIND=ireals),   INTENT (IN)    :: long(ie_tot,je_tot), lati(ie_tot,je_tot)
   REAL (KIND=ireals),   INTENT (OUT)   :: surface(ie_tot,je_tot)
   ! Local variables:
   REAL (KIND=ireals), DIMENSION(ie_tot,je_tot) ::  dx_src, dy_src, source_dx, source_dy

   INTEGER (KIND=iintegers)  ::  i, j

   source_dx(:,:) = 0.0
   source_dy(:,:) = 0.0
   dx_src(:,:) = 0.0
   dy_src(:,:) = 0.0


   DO i = 1 , ie_tot - 1
      dx_src(i,:) = r_earth*sqrt((cos(lati(i,:))*(long(i+1,:)-long(i,:)))**2 + &
                     (lati(i+1,:)-lati(i,:))**2)
   ENDDO
      
   source_dx(1:ie_tot-1,1:je_tot)=dx_src(1:ie_tot-1,1:je_tot)
   source_dx(ie_tot,:)=2*source_dx(ie_tot-1,:)-source_dx(ie_tot-2,:)

   DO j = 1 , je_tot - 1
      dy_src(:,j) = r_earth*sqrt( ( cos(lati(:,j))*(long(:,j+1)-long(:,j)))**2 + &
                     (lati(:,j+1)-lati(:,j))**2)
   ENDDO
   source_dy(1:ie_tot,1:je_tot-1)=dy_src(1:ie_tot,1:je_tot-1)
   source_dy(:,je_tot) = 2*source_dy(:,je_tot-1)-source_dy(:,je_tot-2)


   surface = source_dx * source_dy

   END SUBROUTINE calculate_area
#endif
